window.Gleam.initSearch([{"doc":"singularity","title":"singularity","content":"# singularity\n\n[![Package Version](https://img.shields.io/hexpm/v/singularity)](https://hex.pm/packages/singularity)\n[![Hex Docs](https://img.shields.io/badge/hex-docs-ffaff3)](https://hexdocs.pm/singularity/)\n\nSingularity is a registry for shared singleton actors, in Gleam.\n\n## Purpose\n\nAn actor registry is a system for keeping track of actors over time.  A registry\nis most useful when combined with OTP supervision, as an actor will have a\ndifferent PID and [Subject] after being restarted.  Messages sent to the old\nSubject will yield no response.\n\nBy registering your actors within the supervisor [worker start] function,\nSingularity will be kept up to date during actor restarts.\n\nSingularity is designed to register singleton actors, in other words, a fixed\nnumber of actors, each of which may have a different message type.  If you are\nlooking to manage many actors of the same message type, [chip] is a better\nsolution.\n\n## Inspiration &amp; Alternatives\n\nErlang has a built in registry, but it is not type safe, which makes it\ndifficult to use from Gleam.  The Erlang registry is also shared by all\nprocesses in the BEAM VM, which can make unit tests brittle.\n\nGleam alternatives:\n\n- [chip]\n\n## Operation\n\nSingularity is itself an actor, so you must start it and keep a reference to\nit anywhere you wish to register or retrieve other actors.\n\n```gleam\nlet assert Ok(registry) = singularity.start()\n```\n\nType safety is ensured by utilizing a wrapper type that can hold any of the\nregistered actors, for example:\n\n```gleam\ntype Actor {\n    LightSwitch(Subject(switch.Message))\n    LightBulbA(Subject(bulb.Message))\n    LightBulbB(Subject(bulb.Message))\n}\n```\n\nWhen storing or retrieving an actor, the relevant constructor for the wrapper\ntype is used as a key:\n\n```gleam\n// Registration.\nlet assert Ok(bulb_a) = bulb.start()\nsingularity.register(with: registry, key: LightBulbA, subject: bulb_a)\n\n// Retrieval.\nlet assert BulbA(bulb_a) =\n  singularity.require(from: registry, key: LightBulbA, timeout_ms: 1000)\n```\n\nWhen an actor&#39;s process exits, it will be removed from the registry\nautomatically.  Fetching an actor using the [require](#require) function\nwill block until that actor has been registered.  These two properties\nallow OTP supervisors to start your actors only when their dependencies\nare ready.\n\nWhen using Singularity in a request oriented service (i.e. a web API), you will\ntypically want to store a reference to singularity directly in your request\ncontext custom type.  This allows your request handler functions to fetch the\ndependencies they need during a request.\n\nFor scenarios with high request volumes, or a large number of dependencies,\nyou may prefer to have a supervisor construct your context with all dependencies\nin place.  This will typically require your context, top-level handler, and\nentire HTTP server stack (i.e. mist + wisp), to be created by a supervisor.\n\n## Usage\n\n```sh\ngleam add gleam_erlang\ngleam add gleam_otp\ngleam add singularity\n```\n\n### Basic Example\n\nThis example should compile and makes a good starting point for experimentation.\n\n```gleam\nimport gleam/erlang/process.{type Subject}\nimport gleam/otp/actor\nimport singularity\n\ntype MsgA\n\ntype MsgB\n\ntype Actor {\n  ActorA(Subject(MsgA))\n  ActorB(Subject(MsgB))\n}\n\npub fn main() {\n  let assert Ok(registry) = singularity.start()\n\n  // Create a couple dummy actors, having unique message types.\n  let assert Ok(actor_a) =\n    actor.start(Nil, fn(_msg: MsgA, state) { actor.continue(state) })\n  let assert Ok(actor_b) =\n    actor.start(Nil, fn(_msg: MsgB, state) { actor.continue(state) })\n\n  // Register the actors specifying the wrapper (`Actor`) variant.\n  // Note that the `subject` argument is not wrapped in the Actors\n  // type here.\n  singularity.register(with: registry, key: ActorA, subject: actor_a)\n  singularity.register(with: registry, key: ActorB, subject: actor_b)\n\n  // Retrieve and verify registered actors.  These are wrapped.\n  let assert ActorA(got_a) =\n    singularity.require(from: registry, key: ActorA, timeout_ms: 1000)\n  let assert ActorB(got_b) =\n    singularity.require(from: registry, key: ActorB, timeout_ms: 1000)\n\n  Nil\n}\n```\n\n## Supervisor Example\n\nThis is an abstract example to illustrate usage within a supervisor.\n\n```gleam\nimport gleam/erlang/process\nimport gleam/otp/supervisor\nimport gleam/result\nimport singularity\nimport inner_actor\nimport outer_actor\n\npub type Actor {\n  Inner(inner_actor.Message)\n  Outer(outer_actor.Message)\n}\n\npub fn main() {\n  let assert Ok(registry) = singularity.start()\n\n  // Create worker init functions.\n  let inner_child =\n    supervisor.worker(fn(_state) {\n      // Inner has no dependencies, start and register it.\n      inner_actor.start()\n      |&gt; result.map(singularity.register(registry, Inner, subject: _))\n    })\n\n  let outer_child =\n    supervisor.worker(fn(_state) {\n      // Outer depends on Inner, fetch it.\n      let assert Inner(inner) =\n        singularity.require(registry, Inner, timeout_ms: 5000)\n\n      // Start and register Outer.\n      outer_actor.start(inner)\n      |&gt; result.map(singularity.register(registry, Outer, subject: _))\n    })\n\n  // Start the supervisor.\n  let assert Ok(_) =\n    supervisor.start_spec(\n      supervisor.Spec(\n        argument: Nil,\n        frequency_period: 1,\n        max_frequency: 5,\n        init: fn(children) {\n          children\n          |&gt; supervisor.add(inner_child)\n          |&gt; supervisor.add(outer_child)\n        },\n      ),\n    )\n\n  let assert Outer(outer) =\n    singularity.require(registry, Outer, timeout_ms: 5000)\n\n  outer_actor.do_something(outer)\n}\n```\n\nSee the [examples](https://github.com/jhillyerd/singularity/tree/main/examples)\ndirectory for more complete examples.\n\nFurther documentation can be found at &lt;https://hexdocs.pm/singularity&gt;.\n\n## Development\n\n```sh\ngleam test  # Run the tests\ngleam shell # Run an Erlang shell\n```\n\n\n[chip]:            https://hexdocs.pm/chip/\n[Subject]:         https://hexdocs.pm/gleam_erlang/gleam/erlang/process.html#Subject\n[worker start]:    https://hexdocs.pm/gleam_otp/gleam/otp/supervisor.html#worker \n","url":"index.html"},{"doc":"singularity","title":"Message","content":"pub opaque type Message(wrap)\n Singularity&#39;s message/Subject type.\n\n\nSynonyms:\nsingularity.Message\nsingularity Message","url":"singularity.html#Message"},{"doc":"singularity","title":"register","content":"pub fn register(\n  with actor: Subject(Message(a)),\n  key variant: fn(Subject(b)) -&gt; a,\n  subject subj: Subject(b),\n) -&gt; Subject(b)\n Registers an actor, using the actors wrapper type constructor as a key.\n\n Registered actor processes will be monitored.  Processes that exit will\n be removed from the registry automatically.\n\n Returns the subject unchanged, for ease of use in pipelines.\nSynonyms:\nsingularity.register\nsingularity register","url":"singularity.html#register"},{"doc":"singularity","title":"require","content":"pub fn require(\n  from actor: Subject(Message(a)),\n  key variant: fn(Subject(b)) -&gt; a,\n  timeout_ms timeout: Int,\n) -&gt; a\n Retrieves an actor, using the actors wrapper type constructor as a key.  If\n the actor is not present in the registry, the request will be retried\n every 100ms for a total of `timeout_ms` before crashing.\n\n If your actors need to establish network connections during startup,\n consider using long (5000ms+) timeouts for them and any actors that\n depend on them.\n\nSynonyms:\nsingularity.require\nsingularity require","url":"singularity.html#require"},{"doc":"singularity","title":"start","content":"pub fn start() -&gt; Result(Subject(Message(a)), StartError)\n Starts the registry.\n\n Note: Gleam actors are linked to the parent process by default.\n\nSynonyms:\nsingularity.start\nsingularity start","url":"singularity.html#start"},{"doc":"singularity","title":"stop","content":"pub fn stop(actor: Subject(Message(a))) -&gt; Nil\n Shuts down the registry.  Does not affect registered actors.\n\nSynonyms:\nsingularity.stop\nsingularity stop","url":"singularity.html#stop"},{"doc":"singularity","title":"try_get","content":"pub fn try_get(\n  from actor: Subject(Message(a)),\n  key variant: fn(Subject(b)) -&gt; a,\n) -&gt; Result(a, Nil)\n Retrieves an actor, using the actors wrapper type constructor as a key.  If\n the actor is not present in the registry, returns `None`.\n\nSynonyms:\nsingularity.try_get\nsingularity try_get","url":"singularity.html#try_get"},{"doc":"singularity","title":"singularity","content":" Singularity is a registry for shared singleton actors, in Gleam.\n\n When an actor&#39;s process exits, it will be removed from the registry\n automatically.  Fetching an actor using the [require](#require) function\n will block until that actor has been registered.  These two properties\n allow OTP supervisors to start your actors only when their dependencies\n are ready.\n\n ### Example Usage\n\n ```\n import actor_a\n import actor_b\n\n // Define a wrapper type for the actors (Subjects) you want register.\n type Actors {\n   ActorA(Subject(actor_a.Message))\n   ActorB(Subject(actor_b.Message))\n }\n\n pub fn main() {\n   let assert Ok(registry) = singularity.start()\n   let assert Ok(actor_a_subj) = actor_a.start()\n   let assert Ok(actor_b_subj) = actor_b.start()\n\n   // Register the actors specifying your wrapper (`Actors`) variant.\n   singularity.register(registry, ActorA, actor_a_subj)\n   singularity.register(registry, ActorB, actor_b_subj)\n\n   // Retrieve registered actors, each will be returned via your\n   // wrapper type.\n   let assert ActorA(got_a) =\n     singularity.require(registry, ActorA, timeout_ms: 1000)\n   let assert ActorB(got_b) =\n     singularity.require(registry, ActorB, timeout_ms: 1000)\n }\n ```\n\n The example above is for illustrative purposes only; it will not compile.\n Please see the README and tests for concrete examples.","url":"singularity.html"}]);